#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_conf.h"

typedef enum
{
    START, INSECTION, POSTKEY, POSTVALUE
} PARSESTATE;

static inline HRESULT parseSection (NRTConfig *, char *);
static inline HRESULT parseKey (NRTConfig *, char *);
static inline HRESULT parseValue (NRTConfig *, char *);

#define LINESIZE 200
#define TRUE 1
#define FALSE 0

SO_PUBLIC HRESULT readConfig(NRTConfig *nrtconfig, const char *configfile)
{
    int fd;
    char *token;
    unsigned lineno = 1;
    struct stat sb;
    int numread;

    HRESULT result;
    PARSESTATE state = START;

    fd = open(configfile, O_RDONLY);

    if (fd == -1)
    {
        fprintf(stderr, "Failed to open (%s) in ", configfile);
        perror("readConfig - open");
        exit(-1);
    }

    if (fstat(fd, &sb) == -1)
    {
        perror("fstat");
        exit(-1);
    }

    // Set output nugget options to NULL so we can check to see
    // if they have been declared before attempting to send
    // alert notifications -- molney

    nrtconfig->on1_ip = NULL;
    nrtconfig->on2_ip = NULL;

    char * file = malloc(sb.st_size+1);

    numread = read(fd, file, sb.st_size);
    if (numread < 0)
    {
        printf("\nFailed to read %s: %s\n", configfile, strerror(errno));
        exit(-1);
    }
    close(fd);

    file[sb.st_size] = '\0';

    token = strtok(file, "\n\r\t ");

    while (token != NULL)
    {

        switch (state)
        {
            case START:
                result = parseSection(nrtconfig, token);
                if (result == R_FAIL)
                {
                    printf("\nExpecting a section name on line %u: %s\n", lineno, token);
                    exit(-1);
                }
                state = INSECTION;
                break;

            case INSECTION:
                result = parseKey(nrtconfig, token);
                if (result == R_FAIL)
                {
                    printf("\nExpecting a key name on line %u: %s\n", lineno, token);
                    exit(-1);
                }
                state = POSTKEY;
                break;

            case POSTKEY:
                result = parseValue(nrtconfig, token);
                if (result == R_FAIL)
                {
                    printf("\nExpecting a value on line %u: %s\n", lineno, token);
                    exit(-1);
                }
                state = POSTVALUE;
                break;

            case POSTVALUE:
                result = parseKey(nrtconfig, token);
                if (result == R_FAIL)
                {
                    result = parseSection(nrtconfig, token);
                    if (result == R_FAIL)
                    {
                        printf("\nExpecting a section name or key on line %u: %s\n", lineno, token);
                        exit(-1);
                    }
                    else
                        state = INSECTION;
                }
                else
                    state = POSTKEY;
                break;
        }

        token = strtok(NULL, "\n\r\t ");

    }

    free(file);

    return R_SUCCESS;

}

static inline HRESULT parseSection (NRTConfig *nrtconfig, char *token)
{
    if (!strncasecmp(token, "GLOBAL:", 7))
        nrtconfig->section = GLOBAL;
    else if (!strncasecmp(token, "DISPATCHSRV:", 12))
        nrtconfig->section = DSRV;
    else if (!strncasecmp(token, "NUGGETSRV:", 10))
        nrtconfig->section = NSRV;
    else if (!strncasecmp(token, "DATABASE:", 9))
        nrtconfig->section = DBASE;
    else if (!strncasecmp(token, "OUTPUT:", 7))
        nrtconfig->section = OUTPUTNUG;
    else
        return R_FAIL;
    return R_SUCCESS;
}

static inline HRESULT parseKey (NRTConfig *nrtconfig, char *token)
{
    if (!strncasecmp(token, "MAXTHREADS", 10))
    {
        nrtconfig->currval = &nrtconfig->maximumthreads;
    }
    else if (!strncasecmp(token, "BADMD5CACHESIZE", 15))
    {
        nrtconfig->currval = &nrtconfig->badcachesize;
    }
    else if (!strncasecmp(token, "GOODMD5CACHESIZE", 16))
    {
        nrtconfig->currval = &nrtconfig->goodcachesize;
    }
    else if (!strncasecmp(token, "URLCACHESIZE", 12))
    {
        nrtconfig->currval = &nrtconfig->urlcachesize;
    }
    else if (!strncasecmp(token, "NUGGETTYPE", 10))
    {
        nrtconfig->currval = &nrtconfig->nuggettype;
    }
    else if (!strncasecmp(token, "MACHINES", 8))
    {
        nrtconfig->currval = &nrtconfig->nummachines;
    }
    else if (!strncasecmp(token, "IP", 2))
    {
        nrtconfig->currval = &nrtconfig->dsrvaddr;
    }
    else if (!strncasecmp(token, "PORT", 4))
    {
        nrtconfig->currval = &nrtconfig->dsrvport;
    }
    else if (!strncasecmp(token, "HANDLERDIR", 10))
    {
        nrtconfig->currval = &nrtconfig->handlerdir;
    }
    else if (!strncasecmp(token, "ROUTINGTYPE", 11))
    {
        nrtconfig->currval = &nrtconfig->routingtype;
    }
    else if (!strncasecmp(token, "DB", 2))
    {
        nrtconfig->currval = &nrtconfig->db;
    }
    else if (!strncasecmp(token, "USERNAME", 8))
    {
        nrtconfig->currval = &nrtconfig->username;
    }
    else if (!strncasecmp(token, "PASSWORD", 8))
    {
        nrtconfig->currval = &nrtconfig->password;
    }
    else if (!strncasecmp(token, "HOST", 4))
    {
        nrtconfig->currval = &nrtconfig->dbhost;
    }
    else if (!strncasecmp(token, "NUGADDR", 7))
    {
        nrtconfig->currval = &nrtconfig->nugaddr;
    }
    else if (!strncasecmp(token, "NUGPORT", 7))
    {
        nrtconfig->currval = &nrtconfig->nugport;
    }
    else if (!strncasecmp(token, "NUGNAME", 7))
    {
        nrtconfig->currval = &nrtconfig->nugname;
    }
    else if (!strncasecmp(token, "OUT1_ADDR", 9))
    {
        nrtconfig->currval = &nrtconfig->on1_ip;
    }
    else if (!strncasecmp(token, "OUT1_PORT", 9))
    {
        nrtconfig->currval = &nrtconfig->on1_port;
    }
    else if (!strncasecmp(token, "OUT2_ADDR", 9))
    {
        nrtconfig->currval = &nrtconfig->on2_ip;
    }
    else if (!strncasecmp(token, "OUT2_PORT", 9))
    {
        nrtconfig->currval = &nrtconfig->on2_port;
    }
    else
        return R_FAIL;
    return R_SUCCESS;
}

static inline HRESULT parseValue (NRTConfig *nrtconfig, char *token)
{

    struct in_addr tmp;

    switch (nrtconfig->section)
    {
        case GLOBAL:
            if (nrtconfig->currval == &nrtconfig->maximumthreads)
                nrtconfig->maximumthreads = atoi(token);
            else if (nrtconfig->currval == &nrtconfig->badcachesize)
            {
                nrtconfig->badcachesize = atoi(token);
            }
            else if (nrtconfig->currval == &nrtconfig->goodcachesize)
            {
                nrtconfig->goodcachesize = atoi(token);
            }
            else if (nrtconfig->currval == &nrtconfig->urlcachesize)
            {
                nrtconfig->urlcachesize = atoi(token);
            }
            else if (nrtconfig->currval == &nrtconfig->nuggettype)
            {
                nrtconfig->nuggettype = atoi(token);
            }
            else
                return R_FAIL;
            break;

        case DSRV:
            if (nrtconfig->currval == &nrtconfig->nummachines)
                nrtconfig->nummachines = atoi(token);
            else if (nrtconfig->currval == &nrtconfig->dsrvaddr)
            {
                if (inet_pton(AF_INET, token, &tmp) != 1)
                {
                    printf("\n%s is not a valid ip address\n", token);
                    return R_FAIL;
                }
                else
                {
                    nrtconfig->dsrvaddr = strdup(token);
                }
            }
            else if (nrtconfig->currval == &nrtconfig->dsrvport)
            {
                nrtconfig->dsrvport = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->routingtype)
            {
                if (!strncasecmp(token, "opaque", 6))
                    nrtconfig->routingtype = 0;
                else if (!strncasecmp(token, "transparent", 11))
                    nrtconfig->routingtype = 1;
                else
                    nrtconfig->routingtype = 2;
            }
            else
                return R_FAIL;
            break;

        case NSRV:
            if (nrtconfig->currval == &nrtconfig->handlerdir)
            {
                nrtconfig->handlerdir = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->nugaddr)
            {
                nrtconfig->nugaddr = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->nugport)
            {
                nrtconfig->nugport = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->nugname)
            {
                nrtconfig->nugname = strdup(token);
            }
            else
                return R_FAIL;
            break;

        case DBASE:
            if (nrtconfig->currval == &nrtconfig->db)
            {
                nrtconfig->db = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->dbhost)
            {
                nrtconfig->dbhost = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->username)
            {
                nrtconfig->username = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->password)
            {
                nrtconfig->password = strdup(token);
            }
            else
                return R_FAIL;
            break;

        case OUTPUTNUG:
            if (nrtconfig->currval == &nrtconfig->on1_ip)
            {
                nrtconfig->on1_ip = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->on1_port)
            {
                nrtconfig->on1_port = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->on2_ip)
            {
                nrtconfig->on2_ip = strdup(token);
            }
            else if (nrtconfig->currval == &nrtconfig->on2_port)
            {
                nrtconfig->on2_port = strdup(token);
            }
            else
                return R_FAIL;
            break;

        default:
            return R_FAIL;
    }

    return R_SUCCESS;
}
/*
int main(int argc, char **argv) {
    NRTConfig nrtconfig;
    char *name = "nrt.conf";

    readConfig(&nrtconfig, name);
}
*/
