#include <mysql.h>
#include <syslog.h>
#include <uuid/uuid.h>
#include <arpa/inet.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_db.h"
#include "rzb_dispatcher.h"
#include "rzb_client.h"
#include "rzb_component_db.h"
#include "rzb_intel_db.h"
#include "rzb_intel_global.h"
#include "rzb_management_log.h"

/* SQL STATEMENTS */
static const char NEW_MAIL_SQL[] = "INSERT INTO email_tracker (mail_id, dst_ip, src_ip, timestamp, mail_from, rcpt_to, msg_from, msg_to, subject) VALUES (NULL, INET_ATON('%s'), INET_ATON('%s'), now(), '%s', '%s', '%s', '%s', '%s')";
static const char INSERT_MAIL_DATA[] = "UPDATE email_tracker SET %s = '%s' WHERE mail_id = %u";
static const char INSERT_NAMED_ATTACHMENT[] = "INSERT INTO email_attachments VALUES (%u, %u, '%s', NULL, NULL, NULL)";
static const char INSERT_UNNAMED_ATTACHMENT[] = "INSERT INTO email_attachments VALUES (%u, %u, NULL, NULL, NULL, NULL)";
static const char INSERT_WEB_TRACK[] = "INSERT INTO web_tracker (timestamp, dst_ip, src_ip, url, host, user_agent, cookie, header) VALUES (now(),INET_ATON('%s'), INET_ATON('%s'), '%s', '%s', '%s', '%s', '%s')";
static const char INSERT_DNS_TRACK[] = "INSERT INTO dns_tracker (timestamp, src_ip, dst_ip, tx_id, flags, queries, answers, auth_ns, add_rec) VALUES (now(), INET_ATON('%s'), INET_ATON('%s'), %hu, %hu, '%s', '%s', '%s', '%s')";

HRESULT insertDNSTrackData(DNS_ENTRY *dns_entry, DNS_TRACK *dns_tracker)
{
    MYSQL *conn;
    char *request;
    size_t request_size;

    char *safe_queries;
    char *safe_answers;
    char *safe_auth_ns;
    char *safe_add_rec;
    char src_ip[INET6_ADDRSTRLEN];
    char dst_ip[INET6_ADDRSTRLEN];

    if ((conn = getMYSQLConnection()) == NULL)
        return R_FAIL;

    dst_ip[0] = 0;
    inet_ntop(AF_INET, &dns_entry->dst_ip, dst_ip, sizeof(dst_ip));
    src_ip[0] = 0;
    inet_ntop(AF_INET, &dns_entry->src_ip, src_ip, sizeof(src_ip));

    protectString(dns_entry->queries, dns_tracker->queries_size, &safe_queries);
    protectString(dns_entry->answers, dns_tracker->answers_size, &safe_answers);
    protectString(dns_entry->auth_ns, dns_tracker->auth_ns_size, &safe_auth_ns);
    protectString(dns_entry->add_rec, dns_tracker->add_rec_size, &safe_add_rec);

    request_size = sizeof(INSERT_DNS_TRACK) + (2 * MAX_IP_SIZE) + (2 * MAX_INT_STRING_SIZE);
    request_size += strlen(safe_queries);
    request_size += strlen(safe_answers);
    request_size += strlen(safe_auth_ns);
    request_size += strlen(safe_add_rec);

    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for request", "insderDNSTrackData");
        free(safe_queries);
        free(safe_answers);
        free(safe_auth_ns);
        free(safe_add_rec);
        putMYSQLConnection(conn);
        return R_MALLOC_FAIL;
    }

    snprintf(request, request_size, INSERT_DNS_TRACK, src_ip, dst_ip, dns_entry->tx_id, dns_entry->flags,
             safe_queries, safe_answers, safe_auth_ns, safe_add_rec);

    if (mysql_query(conn, request) != 0)
    {
        submitLog(LOG_ERR, "Failed db query", "insertMailUpdate");
        free(request);
        free(safe_queries);
        free(safe_answers);
        free(safe_auth_ns);
        free(safe_add_rec);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    free(request);
    free(safe_queries);
    free(safe_answers);
    free(safe_auth_ns);
    free(safe_add_rec);
    putMYSQLConnection(conn);
#ifdef RZB_PERF
    pthread_mutex_lock(&perfmutex);
    perf.dns_tracker_insert++;
    pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */
    return R_SUCCESS;

}
HRESULT insertWebTrackData(WEB_FINAL *web_entry, WEB_TRACK *web_tracker)
{
    MYSQL *conn;
    char *request;
    size_t request_size;
    char *safe_url;
    char *safe_host;
    char *safe_ua;
    char *safe_cookie;
    char *safe_header;
    char src_ip[INET6_ADDRSTRLEN];
    char dst_ip[INET6_ADDRSTRLEN];

    if ((conn = getMYSQLConnection()) == NULL)
        return R_FAIL;

    dst_ip[0] = 0;
    inet_ntop(AF_INET, &web_entry->dst_ip, dst_ip, sizeof(dst_ip));
    src_ip[0] = 0;
    inet_ntop(AF_INET, &web_entry->src_ip, src_ip, sizeof(src_ip));

    /* Protect all strings prior to submission */
    if (web_entry->url != NULL)
        protectString(web_entry->url, web_tracker->url_size, &safe_url);
    else
        safe_url = strdup(NULL_ENTRY);
    if (web_entry->host != NULL)
        protectString(web_entry->host, web_tracker->host_size, &safe_host);
    else
        safe_host = strdup(NULL_ENTRY);
    if (web_entry->user_agent != NULL)
        protectString(web_entry->user_agent, web_tracker->user_agent_size, &safe_ua);
    else
        safe_ua = strdup(NULL_ENTRY);
    if (web_entry->cookie != NULL)
        protectString(web_entry->cookie, web_tracker->cookie_size, &safe_cookie);
    else
        safe_cookie = strdup(NULL_ENTRY);
    if (web_entry->header != NULL)
        protectString(web_entry->header, web_tracker->header_size, &safe_header);
    else
        safe_header = strdup(NULL_ENTRY);

    request_size = sizeof(INSERT_WEB_TRACK) + (2 * MAX_IP_SIZE) + MAX_INT_STRING_SIZE;
    request_size += strlen(safe_url);
    request_size += strlen(safe_host);
    request_size += strlen(safe_ua);
    request_size += strlen(safe_cookie);
    request_size += strlen(safe_header);

    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for request", "insertWebTrackData");
        free(safe_url);
        free(safe_host);
        free(safe_ua);
        free(safe_cookie);
        free(safe_header);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    snprintf(request, request_size, INSERT_WEB_TRACK, dst_ip, src_ip, safe_url, safe_host, safe_ua, safe_cookie, safe_header);

    if (mysql_query(conn, request) != 0)
    {
        submitLog(LOG_ERR, "Failed db query", "insertMailUpdate");
        submitLog(LOG_ERR, request, "insertMailUpdate");
        free(safe_url);
        free(safe_host);
        free(safe_ua);
        free(safe_cookie);
        free(safe_header);
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    free(safe_url);
    free(safe_host);
    free(safe_ua);
    free(safe_cookie);
    free(safe_header);
    free(request);
    putMYSQLConnection(conn);
#ifdef RZB_PERF
    pthread_mutex_lock(&perfmutex);
    perf.web_tracker_insert++;
    pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */
    return R_SUCCESS;
}

HRESULT insertAttachment(unsigned mail_id, unsigned component_id, const char *name)
{
    MYSQL *conn;
    char *request;
    size_t request_size;

    if ((conn = getMYSQLConnection()) == NULL)
        return R_FAIL;

    if (name != NULL)
        request_size = sizeof(INSERT_NAMED_ATTACHMENT) + strlen(name) + (2 * MAX_INT_STRING_SIZE);
    else
        request_size = sizeof(INSERT_UNNAMED_ATTACHMENT) + (2 * MAX_INT_STRING_SIZE);


    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate space for request", "insertAttachment");
        putMYSQLConnection(conn);
        return R_MALLOC_FAIL;
    }

    if (name != NULL)
        snprintf(request, request_size, INSERT_NAMED_ATTACHMENT, mail_id, component_id, name);
    else
        snprintf(request, request_size, INSERT_UNNAMED_ATTACHMENT, mail_id, component_id);

    if (mysql_query(conn, request) != 0)
    {
        submitLog(LOG_ERR, "Failed db query", "insertMailUpdate");
        submitLog(LOG_ERR, request, "insertMailUpdate");
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    putMYSQLConnection(conn);
    free(request);
    return R_SUCCESS;
}

HRESULT insertMailUpdate(unsigned mail_id, const char *data, const char *data_type)
{
    MYSQL *conn;
    char *request;
    char *protected_data;
    size_t request_size;

    if ((protectString(data, strlen(data), &protected_data)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to protect string", "insertMailUpdate");
        return R_FAIL;
    }

    request_size = sizeof(INSERT_MAIL_DATA) + MAX_INT_STRING_SIZE;
    request_size += strlen(protected_data);
    request_size += strlen(data_type);
    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate space for request", "insertMailUpdate");
        free(protected_data);
        return R_MALLOC_FAIL;
    }

    snprintf(request, request_size, INSERT_MAIL_DATA, data_type, protected_data, mail_id);
    free(protected_data);

    if ((conn = getMYSQLConnection()) == NULL)
        return R_FAIL;

    if (mysql_query(conn, request) != 0)
    {
        submitLog(LOG_ERR, "Failed db query", "insertMailUpdate");
        submitLog(LOG_ERR, request, "insertMailUpdate");
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    free(request);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

HRESULT insertNewMail(const NEW_MAIL_BLOBS *blobs, const NEW_MAIL *mail, unsigned *mail_id)
{
    MYSQL *conn;
    char *request;
    size_t request_size;
    char *p_mf;
    char *p_rt;
    char *p_su;
    char *p_msg_f;
    char *p_msg_t;
    char src_ip[INET6_ADDRSTRLEN];
    char dst_ip[INET6_ADDRSTRLEN];

    dst_ip[0] = 0;
    inet_ntop(AF_INET, &mail->dst_ip, dst_ip, sizeof(dst_ip));
    src_ip[0] = 0;
    inet_ntop(AF_INET, &mail->src_ip, src_ip, sizeof(src_ip));

    protectString(blobs->mail_from, strlen(blobs->mail_from), &p_mf);
    protectString(blobs->rcpt_to, strlen(blobs->rcpt_to), &p_rt);
    protectString(blobs->subject, strlen(blobs->subject), &p_su);
    protectString(blobs->msg_from, strlen(blobs->msg_from), &p_msg_f);
    protectString(blobs->msg_to, strlen(blobs->msg_to), &p_msg_t);

    /* Request size must be large enough to handle stuff */
    request_size = sizeof(NEW_MAIL_SQL);
    request_size += strlen(p_mf);
    request_size += strlen(p_rt);
    request_size += strlen(p_su);
    request_size += strlen(p_msg_f);
    request_size += strlen(p_msg_t);
    request_size += strlen(src_ip);
    request_size += strlen(dst_ip);

    if ((request = malloc(request_size)) == NULL)
    {
        free(p_mf);
        free(p_rt);
        free(p_su);
        free(p_msg_f);
        free(p_msg_t);
        submitLog(LOG_ERR, "Failed to allocate space for request", "insertNewMail");
        return R_MALLOC_FAIL;
    }

    snprintf(request, request_size, NEW_MAIL_SQL, dst_ip, src_ip, p_mf, p_rt, p_msg_f, p_msg_t, p_su);

    if ((conn = getMYSQLConnection()) == NULL)
        return R_FAIL;

    if (mysql_query(conn, request) != 0)
    {
        submitLog(LOG_ERR, "Failed db query", "insertNewMail");
        submitLog(LOG_ERR, request, "insertNewMail");
        free(request);
        free(p_mf);
        free(p_rt);
        free(p_su);
        free(p_msg_f);
        free(p_msg_t);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }

    *mail_id = mysql_insert_id(conn);

    free(request);
    free(p_mf);
    free(p_rt);
    free(p_su);
    free(p_msg_f);
    free(p_msg_t);
    putMYSQLConnection(conn);
#ifdef RZB_PERF
    pthread_mutex_lock(&perfmutex);
    perf.smtp_tracker_insert++;
    pthread_mutex_unlock(&perfmutex);
#endif /* RZB_PERF */

    return R_SUCCESS;
}

